@@ -16,9 +16,8 @@ const char *phases::getPhaseName(ID Id) {
   switch (Id) {
   case Preprocess: return "preprocessor";
   case Precompile: return "precompiler";
+  case FortranFrontend: return "fortranfrontend";
   case Compile: return "compiler";
-  case FortranFrontend:
-    return "fortran-frontend";
   case Backend: return "backend";
   case Assemble: return "assembler";
   case Link: return "linker";
