@@ -232,12 +232,6 @@ const char *amdgpu::dlr::getLinkCommandArgs(
       LibSuffix.append("/asan");
   }
 
-  // If device debugging turned on, add specially built bc files
-  StringRef libpath = Args.MakeArgString(C.getDriver().Dir + "/../" + LibSuffix);
-  std::string lib_debug_perf_path = FindDebugPerfInLibraryPath(LibSuffix);
-  if (!lib_debug_perf_path.empty())
-    libpath = lib_debug_perf_path;
-
   llvm::SmallVector<std::string, 12> BCLibs;
 
   std::string AsanRTL;
@@ -258,14 +252,10 @@ const char *amdgpu::dlr::getLinkCommandArgs(
   // the look-up of the libomptarget bc lib to happen and if not present
   // where it is expected it means we are using the build tree compiler
   // not the installed compiler.
-  std::string LibDeviceName = "/libomptarget-amdgpu-" + GPUArch.str() + ".bc";
+  std::string LibDeviceName = "/libomptarget-amdgpu.bc";
 
-  SmallString<128> Path(Args.MakeArgString(libpath + LibDeviceName));
-  if (LibSuffix != "lib" || llvm::sys::fs::exists(Path)) {
-    BCLibs.push_back(Args.MakeArgString(Path));
-  } else {
-    // Check if the device library can be found in
-    // one of the LIBRARY_PATH directories.
+  if (!Args.hasArg(options::OPT_nogpulib)) {
+    // Check if libomptarget device bitcode can be found in a LIBRARY_PATH dir
     bool EnvOmpLibDeviceFound = false;
     for (auto &EnvLibraryPath : EnvironmentLibraryPaths) {
       std::string EnvOmpLibDevice = EnvLibraryPath + LibDeviceName;
@@ -275,16 +265,24 @@ const char *amdgpu::dlr::getLinkCommandArgs(
         break;
       }
     }
-    // If LIBRARY_PATH doesn't point to the device library,
-    // then use the default one.
-    if (!EnvOmpLibDeviceFound) {
-      std::string RtDir = "/../runtimes/runtimes-bins/offload";
-      BCLibs.push_back(Args.MakeArgString(libpath + RtDir + LibDeviceName));
-    }
-  }
 
-  if (!AsanRTL.empty()) {
-    if (!Args.hasArg(options::OPT_nogpulib)) {
+    // If not found in LIBRARY_PATH, use default for the correct LibSuffix.
+    if (!EnvOmpLibDeviceFound) {
+      StringRef bc_file_suf = Args.MakeArgString(C.getDriver().Dir + "/../" +
+                                                 LibSuffix + LibDeviceName);
+      StringRef bc_file_lib =
+          Args.MakeArgString(C.getDriver().Dir + "/../lib" + LibDeviceName);
+      if (llvm::sys::fs::exists(bc_file_suf))
+        BCLibs.push_back(Args.MakeArgString(bc_file_suf));
+      else if (llvm::sys::fs::exists(bc_file_lib))
+        // In case a LibSuffix version not found, use suffix "lib"
+        BCLibs.push_back(Args.MakeArgString(bc_file_lib));
+      else
+        TC.getDriver().Diag(diag::err_drv_omp_offload_target_bcruntime_not_found)
+          << "libomptarget-amdgpu.bc";
+    }
+
+    if (!AsanRTL.empty()) {
       // asanrtl is dependent on ockl so for every asanrtl bitcode linking
       // requires ockl but viceversa is not true.
       std::string OcklRTL(RocmInstallation.getOCKLPath());
@@ -293,12 +291,12 @@ const char *amdgpu::dlr::getLinkCommandArgs(
       else
         BCLibs.push_back(OcklRTL);
     }
-  }
 
-  // Add the generic set of libraries, OpenMP subset only
-  BCLibs.append(amdgpu::dlr::getCommonDeviceLibNames(
-      C.getArgs(), C.getDriver(), GPUArch.str(), /* isOpenMP=*/true,
-      RocmInstallation));
+    // Add the generic set of libraries, OpenMP subset only
+    BCLibs.append(amdgpu::dlr::getCommonDeviceLibNames(
+        C.getArgs(), C.getDriver(), GPUArch.str(), /* isOpenMP=*/true,
+        RocmInstallation));
+  }
 
   llvm::for_each(BCLibs, [&](StringRef BCFile) {
     LastLinkArgs.push_back(Args.MakeArgString(BCFile));
@@ -388,7 +386,7 @@ const char *amdgpu::dlr::getLldCommandArgs(
 AMDGPUOpenMPToolChain::AMDGPUOpenMPToolChain(const Driver &D, const llvm::Triple &Triple,
                              const ToolChain &HostTC, const ArgList &Args,
                              const Action::OffloadKind OK)
-    : ROCMToolChain(D, Triple, Args), HostTC(HostTC), OK(OK) {
+    : ROCMToolChain(D, Triple, Args), HostTC(HostTC) {
   // Lookup binaries into the driver directory, this is used to
   // discover the 'amdgpu-arch' executable.
   getProgramPaths().push_back(getDriver().Dir);
@@ -400,10 +398,9 @@ void AMDGPUOpenMPToolChain::addClangTargetOptions(
   HostTC.addClangTargetOptions(DriverArgs, CC1Args, DeviceOffloadingKind);
 
   StringRef GPUArch = DriverArgs.getLastArgValue(options::OPT_march_EQ);
-  assert(!GPUArch.empty() && "Must have an explicit GPU arch.");
 
-  CC1Args.push_back("-target-cpu");
-  CC1Args.push_back(DriverArgs.MakeArgStringRef(GPUArch));
+  assert(DeviceOffloadingKind == Action::OFK_OpenMP &&
+         "Only OpenMP offloading kinds are supported.");
 
   // Extract all the -m options
   std::vector<llvm::StringRef> Features;
@@ -435,8 +432,6 @@ void AMDGPUOpenMPToolChain::addClangTargetOptions(
                          options::OPT_fno_gpu_allow_device_init, false))
     CC1Args.push_back("-fgpu-allow-device-init");
 
-  CC1Args.push_back("-fcuda-allow-variadic-functions");
-
   // Default to "hidden" visibility, as object level linking will not be
   // supported for the foreseeable future.
   if (!DriverArgs.hasArg(options::OPT_fvisibility_EQ,
@@ -463,7 +458,7 @@ void AMDGPUOpenMPToolChain::addClangTargetOptions(
     LibraryPaths.push_back(DriverArgs.MakeArgString(Path));
 
   // Link the bitcode library late if we're using device LTO.
-  if (getDriver().isUsingLTO(/* IsOffload */ true))
+  if (getDriver().isUsingOffloadLTO())
     return;
 
   std::string BitcodeSuffix;
@@ -547,7 +542,7 @@ void AMDGPUOpenMPToolChain::AddClangSystemIncludeArgs(const ArgList &DriverArgs,
   CC1Args.push_back("-internal-isystem");
   CC1Args.push_back(DriverArgs.MakeArgString(D.Dir + "/../include"));
   CC1Args.push_back("-internal-isystem");
-  CC1Args.push_back(DriverArgs.MakeArgString(D.Dir + "/../../include"));
+  CC1Args.push_back(DriverArgs.MakeArgString(D.Dir + "/../../../include"));
 
   HostTC.AddClangSystemIncludeArgs(DriverArgs, CC1Args);
 
@@ -557,6 +552,11 @@ void AMDGPUOpenMPToolChain::AddClangSystemIncludeArgs(const ArgList &DriverArgs,
   CC1Args.push_back(DriverArgs.MakeArgString(P));
 }
 
+void AMDGPUOpenMPToolChain::AddClangCXXStdlibIncludeArgs(
+    const llvm::opt::ArgList &Args, llvm::opt::ArgStringList &CC1Args) const {
+  HostTC.AddClangCXXStdlibIncludeArgs(Args, CC1Args);
+}
+
 /// Convert path list to Fortran frontend argument
 static void AddFlangSysIncludeArg(const ArgList &DriverArgs,
                                   ArgStringList &Flang1args,
@@ -598,12 +598,6 @@ void AMDGPUOpenMPToolChain::AddFlangSystemIncludeArgs(const ArgList &DriverArgs,
   return;
 }
 
-
-void AMDGPUOpenMPToolChain::AddClangCXXStdlibIncludeArgs(const ArgList &Args,
-                                                 ArgStringList &CC1Args) const {
-  HostTC.AddClangCXXStdlibIncludeArgs(Args, CC1Args);
-}
-
 void AMDGPUOpenMPToolChain::AddIAMCUIncludeArgs(const ArgList &Args,
                                                 ArgStringList &CC1Args) const {
   HostTC.AddIAMCUIncludeArgs(Args, CC1Args);
